import re
import fnmatch

_PROTO_RE_STR = '(?P<protocol>\w+)\://'
_HOST_RE_STR = '?P<host>([\w\-_]+\.)*[\w\-_]+|\*'
_PORT_RE_STR = '\:(?P<port>\d{1,5})'

URL_RE = re.compile('({protocol})?({host})?({port})?'.format(protocol=_PROTO_RE_STR,
                                                             host=_HOST_RE_STR,
                                                             port=_PORT_RE_STR))


def sanitize_url(url, protocol=None, host=None, port=None):
    match = URL_RE.match(url)
    if match is None:
        raise ValueError('Invalid URL: {!r}'.format(url))
    pars = match.groupdict()
    _protocol, _host, _port = pars['protocol'], pars['host'], pars['port']
    protocol = protocol if _protocol is None else _protocol
    host = host if _host is None else _host
    port = port if _port is None else _port
    protocol = '' if protocol is None else (protocol + '://')
    port = '' if port is None else ':' + str(port)
    return dict(url='{}{}{}'.format(protocol, host, port),
                protocol=protocol, host=host, port=port)


def filter_patterns(names, patterns):
    patterns = ['*:{}'.format(p) if ':' not in p and '*' not in p else p
                for p in patterns]
    result = set()
    sets = (fnmatch.filter(names, pattern) for pattern in patterns)
    map(result.update, sets)
    return result


def load_config(config_file):
    parser = SafeConfigParser()
    parser.read(config_file)
    dft_global = dict(name='multivisor')

    supervisors = {}
    config = dict(dft_global, supervisors=supervisors)
    config.update(parser.items('global'))
    tasks = []
    for section in parser.sections():
        if not section.startswith('supervisor:'):
            continue
        name = section[len('supervisor:'):]
        section_items = dict(parser.items(section))
        url = section_items.get('url', '')
        supervisors[name] = Supervisor(name, url)
    return config




import sys
import logging
import xmlrpclib

from os import environ
from functools import partial

from gevent import spawn
from gevent.lock import RLock
from gevent.queue import Queue
from zerorpc import stream, Server, LostRemote
from supervisor.childutils import getRPCInterface


READY = 'READY\n'
ACKNOWLEDGED = 'RESULT 2\nOK'

DEFAULT_BIND = 'tcp://*:9002'


def signal(msg):
    sys.stdout.write(msg)
    sys.stdout.flush()


def wait_for_event():
    header_line = sys.stdin.readline()
    event = dict((x.split(':') for x in header_line.split()))
    payload_str = sys.stdin.read(int(event['len']))
    event['payload'] = dict((x.split(':') for x in payload_str.split()))
    return event


def event_dispatcher_loop(handler):
    while True:
        signal(READY)
        event = wait_for_event()
        handler(event)
        signal(ACKNOWLEDGED)


get_rpc = partial(getRPCInterface, environ)


def build_method(name):
    subsystem_name, func_name = name.split('.', 1)
    fname = '{}_{}'.format(subsystem_name, func_name)
    def method(*args):
        subsystem = getattr(get_rpc(), subsystem_name)
        return getattr(subsystem, func_name)(*args)
    method.__name__ = fname
    return fname, method


class Supervisor(object):

    def __init__(self, channel):
        self.channel = channel
        self.event_channels = set()
        self.rpc = get_rpc()
        for name in self.rpc.system.listMethods():
            setattr(self, *build_method(name))

    @stream
    def event_stream(self):
        logging.info('client connected to stream')
        channel = Queue()
        self.event_channels.add(channel)
        try:
            yield 'First event to trigger connection. Please ignore me!'
            for event in channel:
                yield event
        except LostRemote as e:
            logging.info('remote end of stream disconnected')
        finally:
            self.event_channels.remove(channel)

    def publish_event(self, event):
        name = event['eventname']
        if name.startswith('TICK'):
            return
        event = dict(event)
        if name.startswith('PROCESS_STATE'):
            payload = event['payload']
            pname = "{}:{}".format(payload['groupname'], payload['processname'])
            logging.info('handling %s of %s', name, pname)
            try:
                payload['process'] = self.rpc.supervisor.getProcessInfo(pname)
            except xmlrpclib.Fault:
                # probably supervisor is shutting down
                logging.warn('probably shutting down...')
                return
        for channel in self.event_channels:
            channel.put(event)


def run(bind=DEFAULT_BIND):
    def event_consumer_loop():
        # xmlrpclib is not reentrant. We might have several greenlets accessing
        # supervisor at the same time so we serialize event treatment here
        lock = RLock()
        for event in channel:
            try:
                with lock:
                    supervisor.publish_event(event)
            except:
                logging.exception('Error processing %s', event)
    channel = Queue()
    supervisor = Supervisor(channel)
    spawn(event_consumer_loop)
    spawn(event_dispatcher_loop, channel.put)
    server = Server(supervisor)
    server.bind(bind)
    server.run()


def main(args=None):
    import gevent.monkey
    gevent.monkey.patch_all(thread=False, sys=True)

    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument('--bind', help='bind address',
                        default=DEFAULT_BIND)
    parser.add_argument('--log-level', help='log level', type=str,
                        default='INFO',
                        choices=['DEBUG', 'INFO', 'WARN', 'ERROR'])
    options = parser.parse_args(args)

    log_level = getattr(logging, options.log_level.upper())
    log_fmt = '%(levelname)s %(asctime)-15s %(name)s: %(message)s'
    logging.basicConfig(level=log_level, format=log_fmt)

    bind = options.bind
    if '://' not in bind:
        bind = 'tcp://' + bind
    logging.info('Supervisor: %s', environ['SUPERVISOR_SERVER_URL'])
    run(bind)



import json
import time

import louie
import requests


class Multivisor(object):

    def __init__(self, url):
        self.url = url
        self._status = None
        self.notifications = []

    def stop_processes(self, *names):
        return self.post('/process/stop', data=dict(uid=[','.join(names)]))

    def restart_processes(self, *names):
        return self.post('/process/restart', data=dict(uid=[','.join(names)]))

    @property
    def status(self):
        if self._status is None:
            self._status = self.get_status()
        return self._status

    @staticmethod
    def _update_status_stats(status):
        supervisors, processes = status['supervisors'], status['processes']
        s_stats = dict(running=sum((s['running']
                                    for s in status['supervisors'].itervalues())),
                       total=len(supervisors))
        s_stats['stopped'] = s_stats['total'] - s_stats['running']
        p_stats = dict(running=sum((p['running']
                                    for p in status['processes'].itervalues())),
                       total=len(processes))
        p_stats['stopped'] = p_stats['total'] - p_stats['running']
        stats = dict(supervisors=s_stats, processes=p_stats)
        status['stats'] = stats
        return stats

    def get_status(self):
        status = self.get('/refresh').json()
        # reorganize status per process
        status['processes'] = processes = {}
        for supervisor in status['supervisors'].values():
            processes.update(supervisor['processes'])
        self._update_status_stats(status)
        return status

    def refresh_status(self):
        self._status = None
        return self.status

    def get(self, url, params=None, **kwargs):
        result = requests.get(self.url + url, params=params, **kwargs)
        result.raise_for_status()
        return result

    def post(self, url, data=None, json=None, **kwargs):
        result = requests.post(self.url + url, data=data, json=json, **kwargs)
        result.raise_for_status()
        return result

    def __getitem__(self, item):
        return self.get(item).json()

    def __setitem__(self, item, value):
        self.post(item, data=value)

    def events(self):
        stream = self.get('/stream', stream=True)
        for line in stream.iter_lines():
            if line:
                line = line.decode('utf-8')
                if line.startswith('data:'):
                    line = line[5:]
                    try:
                        yield json.loads(line)
                    except ValueError:
                        print 'error', line

    def run(self):
        for event in self.events():
            status = self.status
            name, payload = event['event'], event['payload']
            if name == 'process_changed':
                status['processes'][payload['uid']].update(payload)
                self._update_status_stats(status)
            elif name == 'notification':
                self.notifications.append(payload)
            louie.send(signal=name, sender=self, payload=payload)



from __future__ import absolute_import

import argparse

from .. import util
from . import repl, http


def parse_args(args=None):
    parser = argparse.ArgumentParser()
    parser.add_argument('--url', help='[http://]<host>[:<22000>]',
                        default='localhost:22000')
    return parser.parse_args(args)


def main(args=None):
    import gevent.monkey
    gevent.monkey.patch_all(thread=False)

    options = parse_args(args)
    url = util.sanitize_url(options.url, protocol='http', port=22000)['url']
    multivisor = http.Multivisor(url)
    cli = repl.Repl(multivisor)
    gevent.spawn(multivisor.run)
    cli.run()


if __name__ == '__main__':
    main()



import collections


def group_processes_status_by(processes, group_by='group', filter=None):
    result = collections.defaultdict(lambda : dict(processes={}))
    if filter is None:
        filter = lambda p: True
    for uid, process in processes.items():
        if not filter(process):
            continue
        name = process[group_by]
        order = result[name]
        order['name']  = name
        order['processes'][uid] = process
    return result


def default_process_status(process, max_puid_len=10, group_by='group'):
    nuid = '{{uid:{}}}'.format(max_puid_len).format(uid=process['uid'])
    if group_by in (None, 'process'):
        template = '{nuid} {statename:8} {description}'
    else:
        template = '  {nuid} {statename:8} {description}'
    return template.format(nuid=nuid, **process)


def processes_status(status, group_by='process', filter=None,
                     process_status=default_process_status):
    processes = status['processes']
    puid_len = max(map(len, processes))
    result = []
    if filter is None:
        filter = lambda p: True
    if group_by in (None, 'process'):
        for puid in sorted(processes):
            process = processes[puid]
            if filter(process):
                result.append(process_status(process, max_puid_len=puid_len,
                                             group_by=group_by))
    else:
        grouped = group_processes_status_by(processes, group_by=group_by,
                                            filter=filter)
        for name in sorted(grouped):
            result.append(name + ':')
            for process in grouped[name]['processes'].values():
                result.append(process_status(process, max_puid_len=puid_len,
                                             group_by=group_by))
    return result




#!/usr/bin/env python

import gevent
from gevent.monkey import patch_all
patch_all(thread=False)

import os
import logging

import louie
from gevent import queue, sleep
from gevent.pywsgi import WSGIServer
from flask import Flask, render_template, Response, request, json, jsonify

from ..util import sanitize_url
from ..multivisor import Multivisor


log = logging.getLogger('multivisor')

app = Flask(__name__,
            static_folder='./dist/static',
            template_folder='./dist')


@app.route('/', defaults={'path': ''})
@app.route('/<path:path>')
def catch_all(path):
    return render_template("index.html")


@app.route("/admin/reload")
def reload():
    app.multivisor.reload()
    return 'OK'


@app.route("/refresh")
def refresh():
    app.multivisor.refresh()
    return jsonify(app.multivisor.config)


@app.route("/data")
def data():
    return jsonify(app.multivisor.config)


@app.route("/config/file")
def config_file_content():
    content = app.multivisor.config_file_content
    return jsonify(dict(content=content))


@app.route("/supervisor/update", methods=['POST'])
def update_supervisor():
    names = (unicode.strip(supervisor)
             for supervisor in request.form['supervisor'].split(','))
    app.multivisor.update_supervisors(*names)
    return 'OK'


@app.route("/supervisor/restart", methods=['POST'])
def restart_supervisor():
    names = (unicode.strip(supervisor)
             for supervisor in request.form['supervisor'].split(','))
    app.multivisor.restart_supervisors(*names)
    return 'OK'


@app.route("/supervisor/reread", methods=['POST'])
def reread_supervisor():
    names = (unicode.strip(supervisor)
             for supervisor in request.form['supervisor'].split(','))
    app.multivisor.reread_supervisors(*names)
    return 'OK'


@app.route("/supervisor/shutdown", methods=['POST'])
def shutdown_supervisor():
    names = (unicode.strip(supervisor)
             for supervisor in request.form['supervisor'].split(','))
    app.multivisor.shutdown_supervisors(*names)
    return 'OK'


@app.route("/process/restart", methods=['POST'])
def restart_process():
    patterns = request.form['uid'].split(',')
    procs = app.multivisor.restart_processes(*patterns)
    return 'OK'


@app.route("/process/stop", methods=['POST'])
def stop_process():
    patterns = request.form['uid'].split(',')
    app.multivisor.stop_processes(*patterns)
    return 'OK'


@app.route("/process/list")
def list_processes():
    return jsonify(tuple(app.multivisor.processes.keys()))


@app.route("/process/info/<uid>")
def process_info(uid):
    process = app.multivisor.get_process(uid)
    process.refresh()
    return json.dumps(process)


@app.route("/supervisor/info/<uid>")
def supervisor_info(uid):
    supervisor = app.multivisor.get_supervisor(uid)
    supervisor.refresh()
    return json.dumps(supervisor)


@app.route("/process/log/<stream>/tail/<uid>")
def process_log_tail(stream, uid):
    sname, pname = uid.split(':', 1)
    supervisor = app.multivisor.get_supervisor(sname)
    server = supervisor.server
    if stream == 'out':
        tail = server.supervisor_tailProcessStdoutLog
    else:
        tail = server.supervisor_tailProcessStderrLog
    def event_stream():
        i, offset, length = 0, 0, 2**12
        while True:
            data = tail(pname, offset, length)
            log, offset, overflow = data
            # don't care about overflow in first log message
            if overflow and i:
                length = min(length * 2, 2**14)
            else:
                data = json.dumps(dict(message=log, size=offset))
                yield 'data: {}\n\n'.format(data)
            sleep(1)
            i += 1
    return Response(event_stream(), mimetype="text/event-stream")


@app.route('/stream')
def stream():
    def event_stream():
        client = queue.Queue()
        app.dispatcher.add_listener(client)
        for event in client:
            yield event
        app.dispatcher.remove_listener(client)
    return Response(event_stream(),
                    mimetype="text/event-stream")


class Dispatcher(object):

    def __init__(self):
        self.clients = []
        louie.connect(self.on_multivisor_event, sender='multivisor')

    def add_listener(self, client):
        self.clients.append(client)

    def remove_listener(self, client):
        clients.clients.remove(client)

    def on_multivisor_event(self, signal, payload):
        data = json.dumps(dict(payload=payload, event=signal))
        event = 'data: {0}\n\n'.format(data)
        for client in self.clients:
            client.put(event)


def main(args=None):
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument('--bind', help='[host][:port] (default: 0:22000)',
                        default='0:22000')
    parser.add_argument('-c', help='configuration file',
                        dest='config_file',
                        default='/etc/multivisor.conf')
    parser.add_argument('--log-level', help='log level', type=str,
                        default='INFO',
                        choices=['DEBUG', 'INFO', 'WARN', 'ERROR'])
    options = parser.parse_args(args)

    log_level = getattr(logging, options.log_level.upper())
    log_fmt = '%(levelname)s %(asctime)-15s %(name)s: %(message)s'
    logging.basicConfig(level=log_level, format=log_fmt)

    if not os.path.exists(options.config_file):
        parser.exit(status=2, message='configuration file does not exist. Bailing out!\n')

    bind = sanitize_url(options.bind, host='0', port=22000)['url']

    app.dispatcher = Dispatcher()
    app.multivisor = Multivisor(options)

    http_server = WSGIServer(bind, application=app)
    logging.info('Start accepting requests')
    try:
        http_server.serve_forever()
    except KeyboardInterrupt:
        log.info('Ctrl-C pressed. Bailing out')


if __name__ == "__main__":
    main()



from __future__ import print_function
from __future__ import unicode_literals

import re
import fnmatch
import datetime
import functools

import maya
import louie
from prompt_toolkit import PromptSession, HTML, print_formatted_text
from prompt_toolkit.styles import Style
from prompt_toolkit.history import InMemoryHistory
from prompt_toolkit.completion import WordCompleter
from prompt_toolkit.validation import ValidationError
from prompt_toolkit.application import run_in_terminal
from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.auto_suggest import AutoSuggestFromHistory

from . import util


STYLE = Style.from_dict({
    'stopped': 'ansired',
    'starting': 'ansiblue',
    'running': 'ansigreen',
    'backoff': 'orange',
    'stopping': 'ansiblue',
    'exited': 'ansired bold',
    'fatal': 'violet',
    'unknown': 'grey',
})

NOTIF_STYLE = {
  'DEBUG': 'grey',
  'INFO': 'blue',
  'WARNING': 'orange',
  'ERROR': 'red'
}

def process_description(process):
    # TODO
    status = process['statename']
    if status == 'FATAL':
        return process['description']
    elif process['running']:
        start = maya.MayaDT(process['start'])
        desc = 'pid {pid}, started {start} ({delta})' \
               .format(pid=process['pid'], start=start.rfc2822(), delta=start.slang_time())
    else:
        stop = maya.MayaDT(process['stop'])
        desc = 'stopped on {stop} ({delta} ago)' \
               .format(stop=stop.rfc2822(), delta=stop.slang_time())
    return desc


def process_status(process, max_puid_len=10, group_by='group'):
    state = process['statename']
    uid = '{{uid:{}}}'.format(max_puid_len).format(uid=process['uid'])
    desc = process_description(process)
    text = '{p}{uid} <{lstate}>{state:8}</{lstate}> {description}' \
           .format(p=('' if group_by in (None, 'process') else '  '),
                   uid=uid, state=state, lstate=state.lower(),
                   description=desc)
    return HTML(text)


def processes_status(status, group_by='group', filter='*'):
    filt = lambda p: fnmatch.fnmatch(p['uid'], filter)
    return util.processes_status(status, group_by=group_by, filter=filt,
                                 process_status=process_status)


def print_processes_status(status, *args):
    kwargs = {}
    if args:
        kwargs['filter'] = args[0]
    for text in processes_status(status, **kwargs):
        print_formatted_text(text, style=STYLE)


def cmd(f=None, name=None):
    if f is None:
        return functools.partial(cmd, name=name)
    f.__cmd__ = (name or f.__name__).decode()
    return f


class Commands(object):
    def __init__(self, multivisor):
        self.multivisor = multivisor

    @cmd(name='refresh-status')
    def refresh_status(self):
        """
        refresh              Refresh status (eq of Ctrl+F5 in browser)
        """
        print_processes_status(self.multivisor.refresh_status())

    @cmd
    def status(self, *args):
        """
        status               Status of all processes
        status <pattern>     Status of a pattern of processes
        """
        print_processes_status(self.multivisor.status, *args)

    @cmd
    def restart(self, *args):
        """
        restart <pattern> <pattern>*    Restart a list of process patterns
        """
        if not args:
            raise ValidationError(message='Need at least one process')
        self.multivisor.restart_processes(*args)

    @cmd
    def stop(self, *args):
        """
        stop <pattern> <pattern>*    Stop a list of process patterns
        """
        if not args:
            raise ValidationError(message='Need at least one process')
        self.multivisor.stop_processes(*args)

    @cmd
    def help(self, *args):
        """
        Available commands (type help <topic>):
        =======================================

        {cmds}
        """
        if not args:
            args = 'help',
        cmd = self.get_command(args[0])
        cmds = '  '.join(self.get_commands())
        raw_text = cmd.__doc__.format(cmds=cmds)
        text = '\n'.join(map(unicode.strip, raw_text.split('\n')))
        print_formatted_text(text)

    @classmethod
    def get_commands(cls):
        result = {}
        for name in dir(cls):
            member = getattr(cls, name)
            cmd = getattr(member, '__cmd__', None)
            if cmd:
                result[cmd] = name
        return result

    def get_command(self, name):
        method_name = self.get_commands().get(name)
        if method_name is None:
            raise ValidationError(message="Unknown command '{}'".format(name))
        return getattr(self, method_name)


def Prompt(**kwargs):
    history = InMemoryHistory()
    auto_suggest = AutoSuggestFromHistory()
    prmpt = 'multivisor> '
    return PromptSession(prmpt, history=history, auto_suggest=auto_suggest,
                         **kwargs)


class Repl(object):

    keys = KeyBindings()

    def __init__(self, multivisor):
        self.multivisor = multivisor
        self.commands = Commands(multivisor)
        status = self.multivisor.status
        words = list(status['processes'].keys())
        words.extend(self.commands.get_commands())
        completer = WordCompleter(words)
        self.session = Prompt(completer=completer, bottom_toolbar=self.toolbar,
                              key_bindings=self.keys)
        self.session.app.commands = self.commands
        self.__update_toolbar()
        louie.connect(self.__update_toolbar, sender=self.multivisor)

    def __update_toolbar(self):
        status = self.multivisor.status
        stats = status['stats']
        s_stats, p_stats = stats['supervisors'], stats['processes']
        notifications = self.multivisor.notifications
        if notifications:
            notif = notifications[-1]
        else:
            notif = dict(level='INFO', message='Welcome to multivisor CLI')
        html = '{name} | Supervisors: {s[total]} (' \
               '<b><style bg="green">{s[running]}</style></b>/' \
               '<b><style bg="red">{s[stopped]}</style></b>) ' \
               '| Processes: {p[total]} (' \
               '<b><style bg="green">{p[running]}</style></b>/' \
               '<b><style bg="red">{p[stopped]}</style></b>) ' \
               '| <style bg="{notif_color}">{notif_msg}</style>' \
               .format(name=status['name'], s=s_stats, p=p_stats,
                       notif_color=NOTIF_STYLE[notif['level']],
                       notif_msg=notif['message'])
        self.__toolbar = HTML(html)
        self.session.app.invalidate()

    @keys.add('f5')
    def __on_refresh(event):
        run_in_terminal(event.app.commands.refresh_status())

    def parse_command_line(self, text):
        args = text.split()
        cmd = self.commands.get_command(args[0])
        return cmd, args[1:]

    def run_command_line(self, text):
        try:
            cmd, args = self.parse_command_line(text)
            cmd(*args)
        except KeyboardInterrupt:
            raise
        except Exception as err:
            print_formatted_text(HTML('<red>Error:</red> {}'.format(err)))

    def toolbar(self):
        return self.__toolbar

    def run(self):
        self.commands.status()
        while True:
            try:
                text = self.session.prompt()
                if not text:
                    continue
                self.run_command_line(text)
            except KeyboardInterrupt:
                continue
            except EOFError:
                break




"""
An extension to the standard supervisor RPC interface which subscribes
to internal supervisor events and dispatches them to 0RPC.

disadvantages: it depends on supervisor internal supervisor.events.subscribe
               interface so its usage is quite risky.
advantages: it avoids creating an eventlistener process just to forward events.

The python environment where supervisor runs must have multivisor installed
"""

import os
import queue
import logging
import functools
import threading

from gevent import spawn, hub, sleep
from gevent.queue import Queue
from zerorpc import stream, Server, LostRemote

from supervisor.http import NOT_DONE_YET
from supervisor.rpcinterface import SupervisorNamespaceRPCInterface
from supervisor.events import subscribe, Event, getEventNameByType
# unsubscribe only appears in supervisor > 3.3.4
try:
    from supervisor.events import unsubscribe
except:
    unsubscribe = lambda x, y: None

from .util import sanitize_url


DEFAULT_BIND = 'tcp://*:9002'


def sync(klass):
    def wrap_func(meth):
        @functools.wraps(meth)
        def wrapper(*args, **kwargs):
            args[0]._log.debug('0RPC: called {}'.format(meth.__name__))
            result = meth(*args, **kwargs)
            if callable(result):
                r = NOT_DONE_YET
                while r is NOT_DONE_YET:
                    sleep(0.1)
                    r = result()
                result = r
            return result
        return wrapper

    for name in dir(klass):
        if name.startswith('_') or name == 'event_stream':
            continue
        meth = getattr(klass, name)
        if not callable(meth):
            continue
        setattr(klass, name, wrap_func(meth))
    return klass


# When supervisor is asked to restart, it closes file descriptors
# from 5..1024. Since we are not able to restart the ZeroRPC server
# (see https://github.com/0rpc/zerorpc-python/issues/208) this patch
# prevents supervisor from closing the gevent pipes and 0MQ sockets
# This is a really agressive move but seems to work until the above
# bug is solved
from supervisor.options import ServerOptions
ServerOptions.cleanup_fds = lambda options : None


@sync
class MultivisorNamespaceRPCInterface(SupervisorNamespaceRPCInterface):

    def __init__(self, supervisord, bind):
        SupervisorNamespaceRPCInterface.__init__(self, supervisord)
        self._bind = bind
        self._channel = queue.Queue()
        self._event_channels = set()
        self._server = None
        self._watcher = None
        self._shutting_down = False
        self._log = logging.getLogger('MVRPC')

    def _start(self):
        subscribe(Event, self._handle_event)

    def _stop(self):
        unsubscribe(Event, self._handle_event)
        self._shutting_down = True

    def _shutdown(self):
        # disconnect all streams
        for channel in self._event_channels:
            channel.put(None)
        if self._server is not None:
            self._server.stop()
            self._server.close()

    def _process_event(self, event):
        if self._shutting_down:
            return
        event_name = getEventNameByType(event.__class__)
        stop_event = event_name == 'SUPERVISOR_STATE_CHANGE_STOPPING'
        if stop_event:
            self._log.info('supervisor stopping!')
            self._stop()
        elif event_name.startswith('TICK'):
            return
        try:
            # old supervisor version
            payload_str = event.payload()
        except AttributeError:
            payload_str = str(event)
        payload = dict((x.split(':') for x in payload_str.split()))
        if event_name.startswith('PROCESS_STATE'):
            pname = "{}:{}".format(payload['groupname'], payload['processname'])
            payload['process'] = self.getProcessInfo(pname)
        # broadcast the event to clients
        server = self.supervisord.options.identifier
        new_event = dict(pool='multivisor', server=server,
                         eventname=event_name, payload=payload)
        for channel in self._event_channels:
            channel.put(new_event)
        if stop_event:
            self._shutdown()

    # called on 0RPC server thread
    def _dispatch_event(self):
        while not self._channel.empty():
            event = self._channel.get()
            self._process_event(event)

    # called on main thread
    def _handle_event(self, event):
        if self._server is None:
            reply = start_rpc_server(self, self._bind)
            if isinstance(reply, Exception):
                self._log.critical('severe 0RPC error: %s', reply)
                self._stop()
                self._shutdown()
                return
            self._server, self._watcher = reply
        self._channel.put(event)
        self._watcher.send()

        # handle stop synchronously
        event_name = getEventNameByType(event.__class__)
        if event_name == 'SUPERVISOR_STATE_CHANGE_STOPPING':
            self._server._stop_event.wait()
            self._server = None
            self._watcher = None

    @stream
    def event_stream(self):
        self._log.info('client connected to stream')
        channel = Queue()
        self._event_channels.add(channel)
        try:
            yield 'First event to trigger connection. Please ignore me!'
            for event in channel:
                if event is None:
                    self._log.info('stop: closing client')
                    return
                yield event
        except LostRemote as e:
            self._log.info('remote end of stream disconnected')
        finally:
            self._event_channels.remove(channel)


def start_rpc_server(multivisor, bind):
    future_server = queue.Queue(1)
    th = threading.Thread(target=run_rpc_server, name='RPCServer',
                          args=(multivisor, bind, future_server))
    th.daemon = True
    th.start()
    return future_server.get()


def run_rpc_server(multivisor, bind, future_server):
    multivisor._log.info('0RPC: spawn server on {}...'.format(os.getpid()))
    watcher = hub.get_hub().loop.async()
    stop_event = threading.Event()
    watcher.start(lambda: spawn(multivisor._dispatch_event))
    try:
        server = Server(multivisor)
        server._stop_event = stop_event
        server.bind(bind)
        future_server.put((server, watcher))
        multivisor._log.info('0RPC: server running!')
        server.run()
        multivisor._log.info('0RPC: server stopped!')
    except Exception as err:
        future_server.put(err)
    finally:
        watcher.stop()
        del server
        # prevent reusage of this loop because supervisor closes all ports
        # when a restart happens. It actually doesn't help preventing a crash
        hub.get_hub().destroy(destroy_loop=True)
        multivisor._log.info('0RPC: server thread destroyed!')
    stop_event.set()


def make_rpc_interface(supervisord, bind=DEFAULT_BIND):
    # Uncomment following lines to configure python standard logging
    #log_level = logging.INFO
    #log_fmt = '%(asctime)-15s %(levelname)s %(threadName)-8s %(name)s: %(message)s'
    #logging.basicConfig(level=log_level, format=log_fmt)

    url = sanitize_url(bind, protocol='tcp', host='*', port=9002)
    multivisor = MultivisorNamespaceRPCInterface(supervisord, url['url'])
    multivisor._start()
    return multivisor



from setuptools import setup, find_packages

setup(
    name='multivisor',
    version='5.0.1',
    author='Tiago Coutinho',
    author_email='coutinhotiago@gmail.com',
    description='A centralized supervisor UI (web & CLI)',
    packages=find_packages(),
    package_data={'multivisor.server': ['dist/*',
                                        'dist/static/css/*',
                                        'dist/static/js/*']},
    entry_points=dict(console_scripts=[
        'multivisor=multivisor.server.web:main',
        'multivisor-cli=multivisor.client.cli:main',
        'multivisor-dispatcher=multivisor.dispatcher:main',
        'multivisor-rpc=multivisor.server.zrpc:main']),
    install_requires=['flask', 'gevent', 'supervisor', 'zerorpc', 'louie',
                      'maya', 'requests', 'prompt_toolkit>=2.0.0,<2.1.0'])



#!/usr/bin/env python

import os
import json
import time
import logging
import weakref
from ConfigParser import SafeConfigParser

import louie
import zerorpc
from gevent import queue, spawn, sleep, joinall
from supervisor.xmlrpc import Faults
from supervisor.states import RUNNING_STATES

from .util import sanitize_url, filter_patterns

log = logging.getLogger('multivisor')


class Supervisor(dict):

    Null = {
        'identification': None,
        'api_version': None,
        'version': None,
        'supervisor_version': None,
        'processes': {},
        'running': False,
        'pid': None
    }

    def __init__(self, name, url):
        super(Supervisor, self).__init__(self.Null)
        self.name = self['name'] = name
        self.url = self['url'] = url
        self.log = log.getChild(name)
        addr = sanitize_url(url, protocol='tcp', host=name, port=9002)
        self.address = addr['url']
        self.host = self['host'] = addr['host']
        self.server = zerorpc.Client(self.address)
        # fill supervisor info before events start coming in
        self.event_loop = spawn(self.run)

    def __repr__(self):
        return '{}(name={})'.format(self.__class__.__name__, self.name)

    def __eq__(self, other):
        this, other = dict(self), dict(other)
        this_p = this.pop('processes')
        other_p = other.pop('processes')
        return this == other and this_p.keys() == other_p.keys()

    def run(self):
        last_retry = time.time()
        while True:
            try:
                self.log.info('(re)initializing...')
                self.refresh()
                for i, event in enumerate(self.server.event_stream()):
                    # ignore first event. It serves only to trigger
                    # connection and avoid TimeoutExpired
                    if i != 0:
                        self.handle_event(event)
            except zerorpc.LostRemote:
                self.log.info('Lost remote')
            except zerorpc.TimeoutExpired:
                self.log.info('Timeout expired')
            except Exception as err:
                self.log.info('Connection error')
            finally:
                curr_time = time.time()
                delta = curr_time - last_retry
                if delta < 10:
                    sleep(10 - delta)
                last_retry = time.time()

    def handle_event(self, event):
        name = event['eventname']
        self.log.info('handling %s...', name)
        if name.startswith('SUPERVISOR_STATE'):
            self.refresh()
        elif not self['running']:
            self.refresh()
        elif name.startswith('PROCESS_GROUP'):
            self.refresh()
        elif name.startswith('PROCESS_STATE'):
            payload = event['payload']
            puid = '{}:{}:{}'.format(self.name,
                                     payload['groupname'],
                                     payload['processname'])
            self['processes'][puid].handle_event(event)

    def create_base_info(self):
        return dict(self.Null, name=self.name, url=self.url, host=self.host)

    def read_info(self):
        info = self.create_base_info()
        server = self.server
        info['pid']= pid = server.getPID()
        info['running'] = True
        info['identification'] = server.getIdentification()
        info['api_version'] = server.getAPIVersion()
        info['supervisor_version'] = server.getSupervisorVersion()
        info['processes'] = processes = {}
        for proc in server.getAllProcessInfo():
            process = Process(self, proc)
            processes[process['uid']] = process
        return info

    def update_info(self, info):
        if self == info:
            this_p, info_p = self['processes'], info['processes']
            if this_p != info_p:
                for name, process in info_p.items():
                    if process != this_p[name]:
                        send(process, 'process_changed')
            self.update(info)
        else:
            self.update(info)
            send(self, 'supervisor_changed')

    def refresh(self):
        try:
            info = self.read_info()
        except:
            info = self.create_base_info()
            raise
        finally:
            self.update_info(info)

    def update_server(self, group_names=()):
        server = self.server
        try:
            added, changed, removed = server.reloadConfig()[0]
        except zerorpc.RemoteError as rerr:
            error(rerr.msg)
            return

        # If any gnames are specified we need to verify that they are
        # valid in order to print a useful error message.
        if group_names:
            groups = set()
            for info in server.getAllProcessInfo():
                groups.add(info['group'])
            # New gnames would not currently exist in this set so
            # add those as well.
            groups.update(added)

            for gname in group_names:
                if gname not in groups:
                    self.log.debug('unknown group %s', gname)

        for gname in removed:
            if group_names and gname not in group_names:
                continue
            results = server.stopProcessGroup(gname)
            self.log.debug('stopped process group %s', gname)

            fails = [res for res in results
                     if res['status'] == Faults.FAILED]
            if fails:
                self.log.debug("%s as problems; not removing", gname)
                continue
            server.removeProcessGroup(gname)
            self.log.debug("removed process group %s", gname)

        for gname in changed:
            if group_names and gname not in group_names:
                continue
            server.stopProcessGroup(gname)
            self.log.debug('stopped process group %s', gname)

            server.removeProcessGroup(gname)
            server.addProcessGroup(gname)
            self.log.debug('updated process group %s', gname)

        for gname in added:
            if group_names and gname not in group_names:
                continue
            server.addProcessGroup(gname)
            self.log.debug('added process group %s', gname)

        self.log.info('Updated %s', self.name)

    def _reread(self):
        return self.server.reloadConfig()

    def restart(self):
        # do a reread. If there is an error (bad config) inform the user and
        # and refuse to restart
        try:
            self._reread()
        except zerorpc.RemoteError as rerr:
            error('Cannot restart: {}'.format(rerr.msg))
            return
        result = self.server.restart(timeout=30)
        if result:
            info('Restarted {}'.format(self.name))
        else:
            error('Error restarting {}'.format(self.name))

    def reread(self):
        try:
            added, changed, removed = self._reread()[0]
        except zerorpc.RemoteError as rerr:
            error(rerr.msg)
        else:
            info('Reread config of {} ' \
                            '({} added; {} changed; {} disappeared)'.format(
                            self.name, len(added), len(changed), len(removed)))

    def shutdown(self):
        result = self.server.shutdown()
        if result:
            info('Shut down {}'.format(self.name))
        else:
            error('Error shutting down {}'.format(self.name))


class Process(dict):

    Null = {
        'running': False,
        'pid': None,
        'state': None,
        'statename': 'UNKNOWN'
    }

    def __init__(self, supervisor, *args, **kwargs):
        super(Process, self).__init__(self.Null)
        if args:
            self.update(args[0])
        self.update(kwargs)
        supervisor_name = supervisor['name']
        full_name = self['group'] + ':' + self['name']
        uid = '{}:{}'.format(supervisor_name, full_name)
        self.log = log.getChild(uid)
        self.supervisor = weakref.proxy(supervisor)
        self['full_name'] = full_name
        self['running'] = self['state'] in RUNNING_STATES
        self['supervisor'] = supervisor_name
        self['host'] = supervisor['host']
        self['uid'] = uid

    @property
    def server(self):
        return self.supervisor.server

    @property
    def full_name(self):
        return self['full_name']

    def handle_event(self, event):
        event_name = event['eventname']
        if event_name.startswith('PROCESS_STATE'):
            payload = event['payload']
            proc_info = payload.get('process')
            if proc_info is not None:
                old = self.update_info(proc_info)
                if old != self:
                    old_state, new_state = old['statename'], self['statename']
                    send(self, event='process_changed')
                    if old_state != new_state:
                        info('{} changed from {} to {}'
                                        .format(self, old_state, new_state))

    def read_info(self):
        proc_info = dict(self.Null)
        try:
            proc_info.update(self.server.getProcessInfo(self.full_name))
        except Exception as err:
            self.log.warn('Failed to read info from %s: %s', self['uid'], err)
        return proc_info

    def update_info(self, proc_info):
        old = dict(self)
        proc_info['running'] = proc_info['state'] in RUNNING_STATES
        self.update(proc_info)
        return old

    def refresh(self):
        proc_info = self.read_info()
        self.update_info(proc_info)

    def start(self):
        try:
            self.server.startProcess(self.full_name, False, timeout=30)
        except:
            message = 'Error trying to start {}!'.format(self)
            error(message)
            self.log.exception(message)

    def stop(self):
        try:
            self.server.stopProcess(self.full_name)
        except:
            message = 'Failed to stop {}'.format(self['uid'])
            warning(message)
            self.log.exception(message)

    def restart(self):
        if self['running']:
            self.stop()
        self.start()

    def __str__(self):
        return '{0} on {1}'.format(self['name'], self['supervisor'])

    def __eq__(self, proc):
        p1, p2 = dict(self), dict(proc)
        p1.pop('description')
        p1.pop('now')
        p2.pop('description')
        p2.pop('now')
        return p1 == p2

    def __ne__(self, proc):
        return not self == proc

# Configuration

def load_config(config_file):
    parser = SafeConfigParser()
    parser.read(config_file)
    dft_global = dict(name='multivisor')

    supervisors = {}
    config = dict(dft_global, supervisors=supervisors)
    config.update(parser.items('global'))
    tasks = []
    for section in parser.sections():
        if not section.startswith('supervisor:'):
            continue
        name = section[len('supervisor:'):]
        section_items = dict(parser.items(section))
        url = section_items.get('url', '')
        supervisors[name] = Supervisor(name, url)
    return config


def send(payload, event):
    louie.send(signal=event, sender='multivisor', payload=payload)


def notification(message, level):
    payload = dict(message=message, level=level, time=time.time())
    send(payload, 'notification')


def info(message):
    notification(message, 'INFO')


def warning(message):
    logging.warning(message)
    notification(message, 'WARNING')


def error(message):
    logging.error(message)
    notification(message, 'ERROR')


class Multivisor(object):

    def __init__(self, options):
        self.options = options
        self.reload()

    @property
    def config(self):
        if self._config is None:
            self._config = load_config(self.options.config_file)
        return self._config

    @property
    def config_file_content(self):
        with open(self.options.config_file) as config_file:
            return config_file.read()

    def reload(self):
        self._config = None
        return self.config

    @property
    def supervisors(self):
        return self.config['supervisors']

    @property
    def processes(self):
        procs = (svisor['processes'] for svisor in self.supervisors.values())
        return { puid: proc for sprocs in procs
                 for puid, proc in sprocs.items() }

    def refresh(self):
        tasks = [spawn(supervisor.refresh)
                 for supervisor in self.supervisors.values()]
        joinall(tasks)

    def get_supervisor(self, name):
        return self.supervisors[name]

    def get_process(self, uid):
        supervisor, _ = uid.split(':', 1)
        return self.supervisors[supervisor]['processes'][uid]

    def _do_supervisors(self, operation, *names):
        supervisors = (self.get_supervisor(name) for name in names)
        tasks = [spawn(operation, supervisor) for supervisor in supervisors]
        joinall(tasks)

    def _do_processes(self, operation, *patterns):
        procs = self.processes
        puids = filter_patterns(procs, patterns)
        tasks = [spawn(operation, procs[puid]) for puid in puids]
        joinall(tasks)

    def update_supervisors(self, *names):
        self._do_supervisors(Supervisor.update_server, *names)

    def restart_supervisors(self, *names):
        self._do_supervisors(Supervisor.restart, *names)

    def reread_supervisors(self, *names):
        self._do_supervisors(Supervisor.reread, *names)

    def shutdown_supervisors(self, *names):
        self._do_supervisors(Supervisor.shutdown, *names)

    def restart_processes(self, *patterns):
        self._do_processes(Process.restart, *patterns)

    def stop_processes(self, *patterns):
        self._do_processes(Process.stop, *patterns)
